use std::env;
use std::fs::{DirBuilder, File};
use std::io::{BufRead, BufReader};
use std::path::{Path, PathBuf};
use std::process::Command;

use ::loader::Loader;



/// The default name of the resources directory.
const DEFAULT_RESOURCES_DIR: &'static str = "resources";

/// The default name to be used for the resource manifest file.
const DEFAULT_MANIFEST_FILENAME: &'static str = "resources.msnr";

/// The default name to be used for the resource manifest file.
const DEFAULT_LOADER_FILENAME: &'static str = "resources.msnl";



/// Defines a resource to be compiled.
pub struct Section
{
   /// A unique name used to reference the
   /// resource defined in this Section.
   pub name: String,

   /// A hint for what kind of loader to
   /// use for this resource.
   pub type_hint: String,

   /// The path to the resources src information.
   pub src_path: String,

   /// The path to place the compiled resource at.
   pub dst_path: String,

   /// The compiler to invoke to compile the
   /// resource defined in this Section.
   pub compiler: String,

   /// The arguments to pass to the compiler.
   pub arguments: String
}

/// Processes the "resources.msnr" file and turns it
/// into a "resources.msnl" file.
pub struct Processor
{
   /// The name of the resource manifest file.
   manifest_filename: String,

   /// The path to the resources directory.
   resources_dir: PathBuf,

   /// The path to the output directory.
   output_dir: PathBuf,

   /// The amount of tasks to use when processing the
   /// resources from the manifest file.
   task_count: u64,

   /// The resources sections from the "resources.msnr" file.
   sections: Vec<Section>
}



impl Section
{
   /// Create a new Section with default values.
   pub fn new() -> Section
   {
      Section
      {
         name: String::new(),
         type_hint: String::new(),
         src_path: String::new(),
         dst_path: String::new(),
         compiler: String::new(),
         arguments: String::new()
      }
   }
}

impl Processor
{
   /// Create a new resource manifest Processor.
   pub fn new() -> Processor
   {
      let out_dir: PathBuf;

      // Try to just get the current directory.
      match env::current_dir()
      {
         Ok(dir) =>
         {
            out_dir = PathBuf::from(dir);
         }

         Err(error) =>
         {
            warn!("{}", error);
            out_dir = PathBuf::new();
         }
      }

      Processor
      {
         manifest_filename: String::from(DEFAULT_MANIFEST_FILENAME),
         resources_dir: PathBuf::from(DEFAULT_RESOURCES_DIR),
         output_dir: out_dir,
         task_count: 1u64,
         sections: Vec::new()
      }
   }

   /// Change the directory to load the resource manifest file from.
   /// This is also the directory used as the root when looking for
   /// resources.
   ///
   /// Defaults to "".
   pub fn set_resources_dir(&mut self, dir: &Path)
   {
      self.resources_dir = PathBuf::from(dir);
      println!("Resource dir: {}", self.resources_dir.to_str().unwrap());
   }

   /// Change the directory to place the compiled resources into.
   /// This is also the directory where the loaded manifest file will
   /// be placed.
   ///
   /// Defaults to "resources".
   pub fn set_output_dir(&mut self, dir: &Path)
   {
      self.output_dir = PathBuf::from(dir);
      println!("Output dir: {}", self.output_dir.to_str().unwrap());
   }

   /// Sets the amount of tasks to have live to compile the resources.
   ///
   /// Each resource will be compiled on a seperate blocking thread.
   pub fn set_task_count(&mut self, count: u64)
   {
      self.task_count = count;
      println!("Task count: {}", self.task_count);
   }

   /// Process the resource manifest file into a loader manifest file.
   pub fn process_manifest_file(&mut self)
   {
      // Read the manifest file from the resources directory.
      self.read_manifest_file();

      // Compile all the resources from the manifest file.
      self.compile_resources();

      // Create the loader manifest file.
      self.create_loader_manifest();
   }

   /// Read the manifest file.
   ///
   /// Reading the manifest file can be a blocking
   /// operation since there is nothing else to do until
   /// the contents of this file have been read.
   fn read_manifest_file(&mut self)
   {
      let file: File;
      let mut continue_reading: bool;
      let mut line: String;
      let mut filepath: PathBuf;
      let mut reader: BufReader<File>;
      let mut section_buffer: Vec<String>;

      // Create the filepath for the resource manifest.
      filepath = self.resources_dir.clone();
      filepath.push(self.manifest_filename.clone());

      // Open the manifest file for reading.
      match File::open(filepath)
      {
         Ok(opened_file) =>
         {
            file = opened_file
         }

         Err(error) =>
         {
            error!("{} -- {:?}  {:?}", error,
                   self.resources_dir, self.manifest_filename);
         }
      }

      // Create a reader to read all of the sections of
      // the manifest file.
      reader = BufReader::new(file);

      // Create a buffer to hold the current
      // sections information.
      section_buffer = Vec::new();

      // Read all of the file line by line.
      continue_reading = true;
      while continue_reading == true
      {
         // Read a line from the file.
         line = String::new();
         match reader.read_line(&mut line)
         {
            Ok(num_bytes) =>
            {
               if num_bytes > 0
               {
                  if line == "\n"
                  {
                     // This is an empty line so if there are any
                     // lines in the section buffer then they need
                     // to be read as a section. If not then there
                     // is nothing to do.
                     self.read_manifest_section(&section_buffer);
                     section_buffer = Vec::new();
                  }
                  else
                  {
                     // Add this line to the section buffer.
                     section_buffer.push(line.clone());
                  }
               }
               else
               {
                  // It looks like there is nothing more to be read.
                  continue_reading = false;
               }
            }

            Err(error) =>
            {
               error!("{}", error);
            }
         }
      }

      // Check to make sure that anythin left in
      // the section buffer is handled.
      self.read_manifest_section(&section_buffer);
   }

   /// Read a single manifest section from a given buffer.
   fn read_manifest_section(&mut self, buffer: &Vec<String>)
   {
      let mut header: &str;
      let mut value: &str;
      let mut section: Section;
      let mut values: Vec<&str>;

      // Only do something if there are some
      // lines in the buffer to process.
      if buffer.len() > 0
      {
         // Create a new Section and fill out its information
         // by parsing the data in the buffer.
         section = Section::new();
         for line in buffer.iter()
         {
            // Split each line in the buffer by the '=' character.
            // This should be parsed into 2 lines then.
            values = line.split("=").collect();
            if values.len() == 2
            {
               // Get the header and value to set
               // the Sections values.
               header = values[0].trim();
               value = values[1].trim();
               if header == "name"
               {
                  section.name = String::from(value);
               }
               else if header == "type"
               {
                  section.type_hint = String::from(value);
               }
               else if header == "src"
               {
                  section.src_path = String::from(value);
               }
               else if header == "dst"
               {
                  section.dst_path = String::from(value);
               }
               else if header == "compiler"
               {
                  section.compiler = String::from(value);
               }
               else if header == "args"
               {
                  section.arguments = String::from(value);
               }
            }
            else
            {
               warn!("Detected a name value entry that is incorrect.\n{}",
                     line.trim());
            }
         }

         // Add the newly created section to the list of sections.
         self.sections.push(section);
      }
   }

   /// Compile all the resources in the Section list.
   fn compile_resources(&self)
   {
      let mut dst_path: &Path;
      let mut dst_dir: PathBuf;

      for section in self.sections.iter()
      {
         // Get a copy of the output directory
         // that can be manipulated.
         dst_dir = self.output_dir.clone();

         dst_path = Path::new(&section.dst_path);
         match dst_path.parent()
         {
            Some(dir) =>
            {
               dst_dir.push(dir);
            }

            None =>
            {
            }
         }

         if dst_dir.as_os_str() != ""
         {
            // Create any directory needed for the output
            // of the compilation of the resource in this Section.
            match DirBuilder::new().recursive(true).create(dst_dir.as_path())
            {
               Ok(_) =>
               {
                  debug!("Creating directories for the path:   {:?}",
                         dst_dir);
               }

               Err(error) =>
               {
                  error!("{}", error);
               }
            }
         }

         self.compile_section(section);
      }
   }

   /// Run a Section's compiler for its resource.
   fn compile_section(&self, section: &Section)
   {
      let args: Vec<String>;

      // Build the arguments to pass to the compiler.
      args = self.build_args(section);

      println!("Running compiler {} for resource {} with args '{:?}'.",
               section.compiler, section.name, args);
      if section.compiler.to_lowercase() == "copy"
      {
         self.run_copier(section, &args);
      }
      else
      {
         self.run_compiler(section, &args);
      }
   }

   fn run_copier(&self, section: &Section, _args: &Vec<String>)
   {
      if section.src_path != section.dst_path
      {
         match ::std::fs::copy(Path::new(&section.src_path),
                               Path::new(&section.dst_path))
         {
            Ok(bytes) =>
            {
               debug!("Copied {} bytes from {} to {}.",
                      bytes, section.src_path, section.dst_path);
            }

            Err(error) =>
            {
               error!("{}", error);
            }
         }
      }
   }

   fn run_compiler(&self, section: &Section, args: &Vec<String>)
   {
      let exit_code: i32;
      let out: String;
      let err: String;
      let mut compiler: Command;
      let mut arg_compiler: &mut Command;

      // Create a new command to call the compiler to compile this
      // Section's resource.
      compiler = Command::new(section.compiler.clone());
      arg_compiler = compiler.args(&args);

      // Execute the compiler command.
      match arg_compiler.output()
      {
         Ok(output) =>
         {
            if output.status.success() == true
            {
               // Log that the compiler succeeded.
               info!("{} compiled by {} successfully.",
                     section.name, section.compiler);
            }
            else
            {
               // Get the exit code of the command.
               match output.status.code()
               {
                  Some(code) =>
                  {
                     exit_code = code;
                  }

                  None =>
                  {
                     // Mark that the compiler exited successfully.
                     exit_code = 0;
                  }
               }

               // Get the stdout and stderr strings.
               match String::from_utf8(output.stdout)
               {
                  Ok(command_stdout) =>
                  {
                     out = command_stdout;
                  }

                  Err(error) =>
                  {
                     error!("{}", error);
                  }
               }

               match String::from_utf8(output.stderr)
               {
                  Ok(command_stderr) =>
                  {
                     err = command_stderr;
                  }

                  Err(error) =>
                  {
                     error!("{}", error);
                  }
               }

               // Write a log message that there was an err compiling this
               // Section's resource.
               warn!("{} exited with code {} while compiling {}.\n{}\n{}",
                     section.compiler, exit_code, section.name, out, err);
            }
         }

         Err(error) =>
         {
            error!("{}", error);
         }
      }
   }

   /// Create an array of string arguments to pass
   /// to a compiler based on the given Section.
   fn build_args(&self, section: &Section) -> Vec<String>
   {
      let mut arg: String;
      let mut args: Vec<String>;

      // Create a new array of determined arguments.
      args = Vec::new();

      // First add the input argument. This is a mason compiler must have.
      arg = String::new();
      arg.push_str("-i");
      arg.push_str(" ");
      arg.push_str(section.src_path.as_str());
      args.push(arg);

      // Next add the output argument. This is a mason compiler must have.
      arg = String::new();
      arg.push_str("-o");
      arg.push_str(" ");
      arg.push_str(section.dst_path.as_str());
      args.push(arg);

      // Parse all the arguments to pass to the compiler.
      arg = section.arguments.clone();
      args.push(arg);

      args
   }

   ///
   fn create_loader_manifest(&self)
   {
      let mut loader_filename: PathBuf;
      let mut loader: Loader;

      // Create a new Loader.
      loader = Loader::new();

      // Add all the resource entries to the loader.
      for section in self.sections.iter()
      {
         loader.add_resource_entry(section.name.clone(),
                                   section.type_hint.clone(),
                                   section.dst_path.clone());
      }

      // Create the file to write to.
      loader_filename = self.output_dir.clone();
      loader_filename.push(DEFAULT_LOADER_FILENAME);
      match ::std::fs::File::create(loader_filename.as_path())
      {
         Ok(file) =>
         {
            // Wrtie the loader manifest file.
            match loader.write_to_file(file)
            {
               Ok(_) =>
               {
               }

               Err(error) =>
               {
                  error!("{}", error);
               }
            }
         }

         Err(error) =>
         {
            error!("{}", error);
         }
      }
   }
}
